/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

package uk.co.petertribble.jangle;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Date;
import java.awt.event.*;
import java.math.BigInteger;

import org.jfree.chart.*;
import org.jfree.data.time.*;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.XYPlot;

/**
 * Display snmp data in a graphical chart.
 */
public class SnmpChart implements ActionListener {

    private SnmpController sc;
    private JFreeChart chart;
    private javax.swing.Timer timer;
    private int delay = 30000;
    private int maxage = 1800000;
    private TimeSeriesCollection dataset;
    private boolean showdelta;
    // times in milliseonds
    private long lastsnap;
    private String charttitle;

    private Map <String, TimeSeries> tsmap;
    // save previous values for rates
    private Map <String, BigInteger> valueMap;

    /**
     * Create a new SnmpChart, showing the rate of change of the given
     * statistic.
     *
     * @param sc an SnmpController
     * @param statistic the statistic to show
     * @param charttitle the title to show on the chart
     */
    public SnmpChart(SnmpController sc, String statistic, String charttitle) {
	this(sc, statistic, charttitle, true);
    }

    /**
     * Create a new SnmpChart, showing the given statistic.
     *
     * @param sc an SnmpController
     * @param statistic the statistic to show
     * @param charttitle the title to show on the chart
     * @param showdelta if true, show rates instead of values
     */
    public SnmpChart(SnmpController sc, String statistic, String charttitle,
		boolean showdelta) {
	this.sc = sc;
	this.showdelta = showdelta;
	this.charttitle = charttitle;
	initialize(statistic);
    }

    /**
     * Create a new SnmpChart, showing the rate of change of the given
     * statistics.
     *
     * @param sc an SnmpController
     * @param statistics a List of statistics to show
     * @param showdelta if true, show rates instead of values
     */
    public SnmpChart(SnmpController sc, List <String> statistics,
		boolean showdelta) {
	this(sc, statistics, "", showdelta);
    }

    /**
     * Create a new SnmpChart, showing the rate of change of the given
     * statistics.
     *
     * @param sc an SnmpController
     * @param statistics a List of statistics to show
     * @param charttitle the title to show on the chart
     * @param showdelta if true, show rates instead of values
     */
    public SnmpChart(SnmpController sc, List <String> statistics,
		String charttitle, boolean showdelta) {
	this.sc = sc;
	this.showdelta = showdelta;
	this.charttitle = charttitle;
	initialize(statistics);
    }

    private void initialize(String statistic) {
	List <String> statistics = new ArrayList <String> ();
	statistics.add(statistic);
	initialize(statistics);
    }

    private void initialize(List <String> statistics) {
	tsmap = new HashMap <String, TimeSeries> ();
	valueMap = new HashMap <String, BigInteger> ();
	dataset = new TimeSeriesCollection();
	lastsnap = 0;

	for (String statistic : statistics) {
	    TimeSeries ts = new TimeSeries(statistic);
	    ts.setMaximumItemAge(maxage);
	    dataset.addSeries(ts);
	    tsmap.put(statistic, ts);
	}

	/*
	 * Initialize the saved values so that the first calculated rate
	 * will be the average since statistic creation.
	 */
	for (String statistic : statistics) {
	    valueMap.put(statistic, BigInteger.ZERO);
	}

	updateAccessory();

	String ylabel = showdelta ? SnmpResources.getString("CHART.RATE")
	    : SnmpResources.getString("CHART.VALUE");

	chart = ChartFactory.createTimeSeriesChart(
		charttitle,
		SnmpResources.getString("CHART.TIME"),
		ylabel,
		dataset,
		true,
		true,
		false);

	setAxes();

	startLoop();
    }

    /*
     * Set up the X and Y axes.
     */
    private void setAxes() {
	XYPlot xyplot = chart.getXYPlot();

	String ylabel = showdelta ? SnmpResources.getString("CHART.RATE")
	    : SnmpResources.getString("CHART.VALUE");
	NumberAxis loadaxis = new NumberAxis(ylabel);
	loadaxis.setAutoRange(true);
	loadaxis.setAutoRangeIncludesZero(true);
	xyplot.setRangeAxis(loadaxis);

	DateAxis daxis = new DateAxis(SnmpResources.getString("CHART.TIME"));
	daxis.setAutoRange(true);
	daxis.setFixedAutoRange(maxage);
	xyplot.setDomainAxis(daxis);
    }

    /**
     * Return the chart that is created.
     *
     * @return  The created chart
     */
    public JFreeChart getChart() {
	return chart;
    }

    /**
     * Add another statistic to the current chart. If the statistic is
     * already being charted, it is not added.
     *
     * @param statistic the name of the statistic to add
     */
    public void addStatistic(String statistic) {
	if (!tsmap.containsKey(statistic)) {
	    TimeSeries ts = new TimeSeries(statistic);
	    ts.setMaximumItemAge(maxage);
	    dataset.addSeries(ts);
	    tsmap.put(statistic, ts);
	    /*
	     * Save the current value, not zero. This ensures that the first
	     * rate calculation gives sensible values. It will be low, as it
	     * will average the change between now and the next update over
	     * the whole update interval. This is better than nothing, and
	     * the way to do this properly would be to have separate lastsnap
	     * values saved for each statistic. This would be overkill when
	     * the only affected datapoint is the first one.
	     */
	    // FIXME get the actual value
	    valueMap.put(statistic, BigInteger.ZERO);
	}
    }

    /**
     * Remove the requested statistic from the current chart.
     *
     * @param statistic the name of the statistic to remove
     */
    // FIXME is it valid to remove all statistics?
    public void removeStatistic(String statistic) {
	if (tsmap.containsKey(statistic)) {
	    dataset.removeSeries(tsmap.get(statistic));
	    tsmap.remove(statistic);
	    valueMap.remove(statistic);
	}
    }

    /**
     * Set the maximum age of the chart. Only statistics younger than this
     * age will be shown.
     *
     * @param age The required maximum age in seconds.
     */
    public void setMaxAge(int age) {
	maxage = age*1000;
	// FIXME iterate over the values in the map
	for (String statistic : tsmap.keySet()) {
	    tsmap.get(statistic).setMaximumItemAge(maxage);
	}
    }

    /**
     * Update the statistics.
     */
    public void updateAccessory() {
	double value;
	Millisecond ms = new Millisecond();
	long newsnap = (new Date()).getTime();
	// loop over all statistics
	for (String statistic : tsmap.keySet()) {
	    TimeSeries ts = tsmap.get(statistic);
	    try {
		SnmpObject sno = sc.getValue(statistic);
		BigInteger newvalue = sno.getNumber();
		if (showdelta) {
		    BigInteger lastvalue = valueMap.get(statistic);
		    BigInteger bd = newvalue.subtract(lastvalue);
		    double dt = (double) (newsnap - lastsnap);
		    value = 1000.0*(bd.doubleValue()/dt);
		    valueMap.put(statistic, newvalue);
		} else {
		    value = newvalue.doubleValue();
		}
		ts.add(ms, value);
	    } catch (SnmpException sne) {}
	}
	lastsnap = newsnap;
    }

    /**
     * Start the timer, so the display will continually update.
     */
    public void startLoop() {
	if (timer == null) {
	    timer = new javax.swing.Timer(delay, this);
	}
	timer.start();
    }

    /**
     * Stop the timer, so the display will not update.
     */
    public void stopLoop() {
	if (timer != null) {
	    timer.stop();
	}
    }

    /**
     * Set the update delay.
     *
     * @param interval the delay value in seconds
     */
    public void setDelay(int interval) {
	delay = interval*1000;
	if (timer != null) {
	    timer.setDelay(delay);
	}
    }

    public void actionPerformed(ActionEvent e) {
	updateAccessory();
    }
}
